#![deny(unsafe_code)]
#![no_main]
#![no_std]
#![allow(unused_imports)]

use panic_rtt_target as _panic_handler;

/* declare a submodule for handling tim8 interrupts */
mod tim8;

mod periods {
    pub const LED_BLINK_PERIOD: u32 = 138_888_888_u32;
}

mod led_blinker {
    
    use stm32f4xx_hal::{
        gpio::{Output, PushPull},
    };

    pub struct Blinker {
        is_on: bool,
        led: stm32f4xx_hal::gpio::Pin<Output<PushPull>, 'A', 5>,
    }
    
    // stm32f4xx_hal::gpio::Pin<Output<PushPull>, 'A', 5>
    impl Blinker {
        pub fn init(is_on: bool, led: stm32f4xx_hal::gpio::Pin<Output<PushPull>, 'A', 5>) -> Blinker {
            let mut blinker = Blinker {
                is_on: is_on,
                led: led,
            };
            if blinker.is_on {
                blinker.led.set_high();
            } else {
                blinker.led.set_low();
            }
            return blinker;
        }
        
        pub fn toggle(&mut self) {
            let is_on: &mut bool = &mut self.is_on;
            if *is_on {
                self.led.set_high();
                *is_on = false;
            } else {
                self.led.set_low();
                *is_on = true;
            }
        }
    }
}

/* declare the RTIC application itself */
#[rtic::app(device = stm32f4xx_hal::stm32, peripherals = true, dispatchers = [EXTI0])]
mod app {

    const BLINKER_START_DURATION : rtic::rtic_monotonic::Nanoseconds = rtic::rtic_monotonic::Nanoseconds(LED_BLINK_PERIOD);

    use crate::periods::*;

    /* bring dependencies into scope */
    use rtt_target::{rprintln, rtt_init_print};
    use stm32f4xx_hal::{
        gpio::{gpioc::PC6, Alternate},
        gpio::{gpioa::PA5, Output, PushPull},
        prelude::*,
        pwm_input::PwmInput,
        stm32::TIM8,
        timer::Timer,
    };

    use crate::led_blinker::*;
    
    use dwt_systick_monotonic::DwtSystick;
    use rtic::time::duration::*;

    const MONO_HZ: u32 = 8_000_000; // 8 MHz

    #[monotonic(binds = SysTick, default = true)]
    type MyMono = DwtSystick<MONO_HZ>;


    /// PWM input monitor type
    pub(crate) type PwmMonitor = PwmInput<TIM8, PC6<Alternate<3>>>;

    /* resources shared across RTIC tasks */
    #[shared]
    struct Shared {
        /// the last observed position of the turret
        last_observed_turret_position: f32,
    }

    /* resources local to specific RTIC tasks */
    #[local]
    struct Local {
        monitor: PwmMonitor,
        led_blinker: crate::led_blinker::Blinker,
    }

    #[init]
    fn init(ctx: init::Context) -> (Shared, Local, init::Monotonics) {
        let mut dcb = ctx.core.DCB;
        let dwt = ctx.core.DWT;
        let systick = ctx.core.SYST;
 
        //let mono = DwtSystick::new(&mut dcb, dwt, systick, MONO_HZ);
        let mono = DwtSystick::new(&mut dcb, dwt, systick, MONO_HZ);

        // Enable RTT logging
        rtt_init_print!();
        rprintln!("hello, world!");
        // retrieve the RCC register, which is needed to obtain a handle to the clocks
        let rcc = ctx.device.RCC.constrain();
        // then retrieve the clocks, so we can configure timers later on
        let clocks = rcc.cfgr.freeze();


        // Setup LED
        let gpioa = ctx.device.GPIOA.split();
        let led_green = gpioa
            .pa5.into_push_pull_output();
            
        let led_blinker = crate::led_blinker::Blinker::init(true, led_green);

        // Schedule the blinking task
        
        blinker::spawn_after(BLINKER_START_DURATION).unwrap();


        // Timer stuff

        // obtain a reference to the GPIOC register block, so we can configure pins on the PC bus.
        let gpioc = ctx.device.GPIOC.split();

        // Configure one of TIM8's CH1 pins, so that its attached to the peripheral.
        // We need to do this since the pins are multiplexed across multiple peripherals
        let tim8_cc1 = gpioc.pc6.into_alternate();

        // Configure TIM8 into PWM input mode.
        // This requires a "best guess" of the input frequency in order to be accurate.
        // Note: as a side-effect TIM8's interrupt is enabled and fires whenever a capture-compare
        //      cycle is complete. See the reference manual's paragraphs on PWM Input.
        let monitor = Timer::new(ctx.device.TIM8, &clocks).pwm_input(240.hz(), tim8_cc1);

        // lastly return the shared and local resources, as per RTIC's spec.
        (
            Shared {
                last_observed_turret_position: 0.0,
            },
            Local { 
                monitor: monitor,
                led_blinker: led_blinker,
            },
            init::Monotonics(mono),
        )
    }


    #[task(local = [led_blinker])]
//    #[task]
    fn blinker(ctx: blinker::Context) {
        ctx.local.led_blinker.toggle();
        blinker::spawn_after(BLINKER_START_DURATION).unwrap();
    }

    /* bring tim8's interrupt handler into scope */
    use crate::tim8::tim8_cc;

    // RTIC docs specify we can modularize the code by using these `extern` blocks.
    // This allows us to specify the tasks in other modules and still work within
    // RTIC's infrastructure.
    extern "Rust" {
        #[task(binds=TIM8_CC, local=[monitor], shared=[last_observed_turret_position])]
        fn tim8_cc(context: tim8_cc::Context);
    }
}
